from typing import List, Optional, cast

import h264_profile_level_id as h264

from .rtp_parameters import (
    ExtendedCodec,
    ExtendedHeaderExtension,
    ExtendedRtpCapabilities,
    MediaKind,
    RtcpFeedback,
    RtcpParameters,
    RtpCapabilities,
    RtpCodec,
    RtpCodecCapability,
    RtpCodecParameters,
    RtpEncodingParameters,
    RtpHeaderExtension,
    RtpHeaderExtensionParameters,
    RtpParameters,
)
from .sctp_parameters import NumSctpStreams, SctpCapabilities, SctpStreamParameters


RTP_PROBATOR_MID = "probator"
RTP_PROBATOR_SSRC = 1234
RTP_PROBATOR_CODEC_PAYLOAD_TYPE = 127


def matchCodecs(
    aCodec: RtpCodec, bCodec: RtpCodec, strict: bool = False, modify: bool = False
) -> bool:
    aMimeType = aCodec.mimeType.lower()
    bMimeType = bCodec.mimeType.lower()

    if aMimeType != bMimeType:
        return False
    if aCodec.clockRate != bCodec.clockRate:
        return False
    if aCodec.channels != bCodec.channels:
        return False

    if aMimeType == "video/h264":
        aPacketizationMode = aCodec.parameters.get("packetization-mode", 0)
        bPacketizationMode = bCodec.parameters.get("packetization-mode", 0)

        if aPacketizationMode != bPacketizationMode:
            return False

        if strict:
            if not h264.isSameProfile(aCodec.parameters, bCodec.parameters):
                return False

            if modify:
                try:
                    selectedProfileLevelId = h264.generateProfileLevelIdForAnswer(
                        aCodec.parameters, bCodec.parameters
                    )
                except TypeError:
                    return False
                else:
                    if selectedProfileLevelId:
                        aCodec.parameters["profile-level-id"] = selectedProfileLevelId
                        bCodec.parameters["profile-level-id"] = selectedProfileLevelId
                    else:
                        aCodec.parameters.pop("profile-level-id", None)
                        bCodec.parameters.pop("profile-level-id", None)

    elif aMimeType == "video/vp9" and strict:
        aProfileId = aCodec.parameters.get("profile-id", 0)
        bProfileId = bCodec.parameters.get("profile-id", 0)

        if aProfileId != bProfileId:
            return False

    return True


def isRtxCodec(codec: Optional[RtpCodec]) -> bool:
    if not codec:
        return False

    return codec.mimeType.lower().endswith("/rtx")


def reduceRtcpFeedback(codecA: RtpCodec, codecB: RtpCodec) -> List[RtcpFeedback]:
    reducedRtcpFeedback: List[RtcpFeedback] = []

    for aFb in codecA.rtcpFeedback:
        matchingBFb = next(
            (
                bFb
                for bFb in codecB.rtcpFeedback
                if bFb.type == aFb.type
                and (
                    bFb.parameter == aFb.parameter
                    or (not bFb.parameter and not aFb.parameter)
                )
            ),
            None,
        )

        if matchingBFb:
            reducedRtcpFeedback.append(matchingBFb)

    return reducedRtcpFeedback


def matchHeaderExtensions(aExt: RtpHeaderExtension, bExt: RtpHeaderExtension) -> bool:
    if aExt.kind and bExt.kind and aExt.kind != bExt.kind:
        return False
    if aExt.uri != bExt.uri:
        return False

    return True


def getExtendedRtpCapabilities(
    localCaps: RtpCapabilities,
    remoteCaps: RtpCapabilities,
    preferLocalCodecsOrder: bool = False,
) -> ExtendedRtpCapabilities:
    extendedRtpCapabilities = ExtendedRtpCapabilities()

    if preferLocalCodecsOrder:
        for localCodec in localCaps.codecs:
            if isRtxCodec(localCodec):
                continue

            matchingRemoteCodec = next(
                (
                    remoteCodec
                    for remoteCodec in remoteCaps.codecs
                    if matchCodecs(remoteCodec, localCodec, strict=True, modify=True)
                ),
                None,
            )

            if not matchingRemoteCodec:
                continue

            localPayloadType = localCodec.preferredPayloadType
            remotePayloadType = matchingRemoteCodec.preferredPayloadType

            if localPayloadType is None or remotePayloadType is None:
                continue

            extendedRtpCapabilities.codecs.append(
                ExtendedCodec(
                    kind=localCodec.kind,
                    mimeType=localCodec.mimeType,
                    clockRate=localCodec.clockRate,
                    channels=localCodec.channels,
                    localPayloadType=localPayloadType,
                    remotePayloadType=remotePayloadType,
                    localRtxPayloadType=None,
                    remoteRtxPayloadType=None,
                    localParameters=localCodec.parameters,
                    remoteParameters=matchingRemoteCodec.parameters,
                    rtcpFeedback=reduceRtcpFeedback(localCodec, matchingRemoteCodec),
                )
            )
    else:
        for remoteCodec in remoteCaps.codecs:
            if isRtxCodec(remoteCodec):
                continue

            matchingLocalCodec = next(
                (
                    localCodec
                    for localCodec in localCaps.codecs
                    if matchCodecs(localCodec, remoteCodec, strict=True, modify=True)
                ),
                None,
            )

            if not matchingLocalCodec:
                continue

            localPayloadType = matchingLocalCodec.preferredPayloadType
            remotePayloadType = remoteCodec.preferredPayloadType

            if localPayloadType is None or remotePayloadType is None:
                continue

            extendedRtpCapabilities.codecs.append(
                ExtendedCodec(
                    kind=matchingLocalCodec.kind,
                    mimeType=matchingLocalCodec.mimeType,
                    clockRate=matchingLocalCodec.clockRate,
                    channels=matchingLocalCodec.channels,
                    localPayloadType=localPayloadType,
                    remotePayloadType=remotePayloadType,
                    localRtxPayloadType=None,
                    remoteRtxPayloadType=None,
                    localParameters=matchingLocalCodec.parameters,
                    remoteParameters=remoteCodec.parameters,
                    rtcpFeedback=reduceRtcpFeedback(matchingLocalCodec, remoteCodec),
                )
            )

    for extendedCodec in extendedRtpCapabilities.codecs:
        matchingLocalRtxCodec = next(
            (
                localCodec
                for localCodec in localCaps.codecs
                if isRtxCodec(localCodec)
                and localCodec.parameters.get("apt") == extendedCodec.localPayloadType
            ),
            None,
        )
        matchingRemoteRtxCodec = next(
            (
                remoteCodec
                for remoteCodec in remoteCaps.codecs
                if isRtxCodec(remoteCodec)
                and remoteCodec.parameters.get("apt") == extendedCodec.remotePayloadType
            ),
            None,
        )

        if matchingLocalRtxCodec and matchingRemoteRtxCodec:
            extendedCodec.localRtxPayloadType = (
                matchingLocalRtxCodec.preferredPayloadType
            )
            extendedCodec.remoteRtxPayloadType = (
                matchingRemoteRtxCodec.preferredPayloadType
            )

    for remoteExt in remoteCaps.headerExtensions:
        matchingLocalExt = next(
            (
                localExt
                for localExt in localCaps.headerExtensions
                if matchHeaderExtensions(localExt, remoteExt)
            ),
            None,
        )

        if not matchingLocalExt:
            continue

        extendedExt = ExtendedHeaderExtension(
            kind=remoteExt.kind,
            uri=remoteExt.uri,
            sendId=matchingLocalExt.preferredId,
            recvId=remoteExt.preferredId,
            encrypt=bool(matchingLocalExt.preferredEncrypt),
            direction="sendrecv",
        )

        if remoteExt.direction == "recvonly":
            extendedExt.direction = "sendonly"
        elif remoteExt.direction == "sendonly":
            extendedExt.direction = "recvonly"
        elif remoteExt.direction == "inactive":
            extendedExt.direction = "inactive"

        extendedRtpCapabilities.headerExtensions.append(extendedExt)

    return extendedRtpCapabilities


def getRecvRtpCapabilities(
    extendedRtpCapabilities: ExtendedRtpCapabilities,
) -> RtpCapabilities:
    return _getRtpCapabilities("recvonly", extendedRtpCapabilities)


def getSendRtpCapabilities(
    extendedRtpCapabilities: ExtendedRtpCapabilities,
) -> RtpCapabilities:
    return _getRtpCapabilities("sendonly", extendedRtpCapabilities)


def getSendingRtpParameters(
    kind: MediaKind, extendedRtpCapabilities: ExtendedRtpCapabilities
) -> RtpParameters:
    rtpParameters = RtpParameters(
        mid=None, msid=None, rtcp=RtcpParameters(cname=None, mux=True)
    )

    for extendedCodec in extendedRtpCapabilities.codecs:
        if extendedCodec.kind != kind:
            continue

        rtpParameters.codecs.append(
            RtpCodecParameters(
                mimeType=extendedCodec.mimeType,
                payloadType=extendedCodec.localPayloadType,
                clockRate=extendedCodec.clockRate,
                channels=extendedCodec.channels,
                parameters=extendedCodec.localParameters,
                rtcpFeedback=extendedCodec.rtcpFeedback,
            )
        )

        if extendedCodec.localRtxPayloadType:
            rtpParameters.codecs.append(
                RtpCodecParameters(
                    mimeType=f"{extendedCodec.kind}/rtx",
                    payloadType=extendedCodec.localRtxPayloadType,
                    clockRate=extendedCodec.clockRate,
                    parameters={"apt": extendedCodec.localPayloadType},
                    rtcpFeedback=[],
                )
            )

    for extendedExtension in extendedRtpCapabilities.headerExtensions:
        if (extendedExtension.kind and extendedExtension.kind != kind) or (
            extendedExtension.direction not in ["sendrecv", "sendonly"]
        ):
            continue

        rtpParameters.headerExtensions.append(
            RtpHeaderExtensionParameters(
                uri=extendedExtension.uri,
                id=extendedExtension.sendId,
                encrypt=extendedExtension.encrypt,
                parameters={},
            )
        )

    return rtpParameters


def getSendingRemoteRtpParameters(
    kind: MediaKind, extendedRtpCapabilities: ExtendedRtpCapabilities
) -> RtpParameters:
    rtpParameters = RtpParameters(
        mid=None, msid=None, rtcp=RtcpParameters(cname=None, mux=True)
    )

    for extendedCodec in extendedRtpCapabilities.codecs:
        if extendedCodec.kind != kind:
            continue

        rtpParameters.codecs.append(
            RtpCodecParameters(
                mimeType=extendedCodec.mimeType,
                payloadType=extendedCodec.localPayloadType,
                clockRate=extendedCodec.clockRate,
                channels=extendedCodec.channels,
                parameters=extendedCodec.remoteParameters,
                rtcpFeedback=extendedCodec.rtcpFeedback,
            )
        )

        if extendedCodec.localRtxPayloadType:
            rtpParameters.codecs.append(
                RtpCodecParameters(
                    mimeType=f"{extendedCodec.kind}/rtx",
                    payloadType=extendedCodec.localRtxPayloadType,
                    clockRate=extendedCodec.clockRate,
                    parameters={"apt": extendedCodec.localPayloadType},
                    rtcpFeedback=[],
                )
            )

    for extendedExtension in extendedRtpCapabilities.headerExtensions:
        if (extendedExtension.kind and extendedExtension.kind != kind) or (
            extendedExtension.direction not in ["sendrecv", "sendonly"]
        ):
            continue

        rtpParameters.headerExtensions.append(
            RtpHeaderExtensionParameters(
                uri=extendedExtension.uri,
                id=extendedExtension.sendId,
                encrypt=extendedExtension.encrypt,
                parameters={},
            )
        )

    if any(
        ext.uri
        == "http://www.ietf.org/id/draft-holmer-rmcat-transport-wide-cc-extensions-01"
        for ext in rtpParameters.headerExtensions
    ):
        for codec in rtpParameters.codecs:
            codec.rtcpFeedback = [
                fb for fb in codec.rtcpFeedback if fb.type != "goog-remb"
            ]
    elif any(
        ext.uri == "http://www.webrtc.org/experiments/rtp-hdrext/abs-send-time"
        for ext in rtpParameters.headerExtensions
    ):
        for codec in rtpParameters.codecs:
            codec.rtcpFeedback = [
                fb for fb in codec.rtcpFeedback if fb.type != "transport-cc"
            ]
    else:
        for codec in rtpParameters.codecs:
            codec.rtcpFeedback = [
                fb
                for fb in codec.rtcpFeedback
                if fb.type not in ["transport-cc", "goog-remb"]
            ]

    return rtpParameters


def reduceCodecs(
    codecs: List[RtpCodecParameters], capCodec: Optional[RtpCodecCapability] = None
) -> List[RtpCodecParameters]:
    filteredCodecs: List[RtpCodecParameters] = []

    if not capCodec:
        filteredCodecs.append(codecs[0])

        if len(codecs) > 1 and isRtxCodec(codecs[1]):
            filteredCodecs.append(codecs[1])
    else:
        for idx, codec in enumerate(codecs):
            if matchCodecs(codec, capCodec, strict=True):
                filteredCodecs.append(codec)

                if idx + 1 < len(codecs) and isRtxCodec(codecs[idx + 1]):
                    filteredCodecs.append(codecs[idx + 1])

                break

        if not filteredCodecs:
            raise TypeError("no matching codec found")

    return filteredCodecs


def generateProbatorRtpParameters(videoRtpParameters: RtpParameters) -> RtpParameters:
    videoRtpParameters = videoRtpParameters.model_copy(deep=True)
    validateAndNormalizeRtpParameters(videoRtpParameters)

    rtpParameters = RtpParameters(
        mid=RTP_PROBATOR_MID,
        msid=None,
        encodings=[
            RtpEncodingParameters(
                ssrc=RTP_PROBATOR_SSRC,
                rid=None,
                codecPayloadType=None,
                rtx=None,
                dtx=False,
                scalabilityMode=None,
                scaleResolutionDownBy=None,
                maxBitrate=None,
                maxFramerate=None,
                adaptivePtime=False,
                priority=None,
                networkPriority=None,
            )
        ],
        rtcp=RtcpParameters(cname="probator", mux=True),
    )

    rtpParameters.codecs.append(videoRtpParameters.codecs[0])
    rtpParameters.codecs[0].payloadType = RTP_PROBATOR_CODEC_PAYLOAD_TYPE
    rtpParameters.headerExtensions = videoRtpParameters.headerExtensions

    return rtpParameters


def canSend(kind: MediaKind, extendedRtpCapabilities: ExtendedRtpCapabilities) -> bool:
    return any(codec.kind == kind for codec in extendedRtpCapabilities.codecs)


def canReceive(
    rtpParameters: RtpParameters, extendedRtpCapabilities: ExtendedRtpCapabilities
) -> bool:
    validateAndNormalizeRtpParameters(rtpParameters)

    if not rtpParameters.codecs:
        return False

    firstMediaCodec = rtpParameters.codecs[0]

    return any(
        codec.remotePayloadType == firstMediaCodec.payloadType
        for codec in extendedRtpCapabilities.codecs
    )


def validateAndNormalizeRtpCapabilities(caps: RtpCapabilities) -> None:
    if not isinstance(caps, RtpCapabilities):
        raise TypeError("caps is not an object")

    for codec in caps.codecs:
        validateAndNormalizeRtpCodecCapability(codec)

    for ext in caps.headerExtensions:
        validateAndNormalizeRtpHeaderExtension(ext)


def validateAndNormalizeRtpParameters(params: RtpParameters) -> None:
    if not isinstance(params, RtpParameters):
        raise TypeError("params is not an object")

    if params.mid is not None and not isinstance(params.mid, str):
        raise TypeError("params.mid is not a string")

    for codec in params.codecs:
        validateAndNormalizeRtpCodecParameters(codec)

    for ext in params.headerExtensions:
        validateRtpHeaderExtensionParameters(ext)

    for encoding in params.encodings:
        validateAndNormalizeRtpEncodingParameters(encoding)

    if params.rtcp is None:
        params.rtcp = RtcpParameters(cname=None, mux=True)
    elif not isinstance(params.rtcp, RtcpParameters):
        raise TypeError("params.rtcp is not an object")

    validateAndNormalizeRtcpParameters(params.rtcp)


def validateAndNormalizeSctpStreamParameters(params: SctpStreamParameters) -> None:
    if not isinstance(params, SctpStreamParameters):
        raise TypeError("params is not an object")

    if params.streamId is None or not isinstance(params.streamId, int):
        raise TypeError("missing params.streamId")

    if params.ordered is None:
        params.ordered = True
    elif not isinstance(params.ordered, bool):
        raise TypeError("invalid params.ordered")

    if params.maxPacketLifeTime is not None and not isinstance(
        params.maxPacketLifeTime, int
    ):
        raise TypeError("invalid params.maxPacketLifeTime")

    if params.maxRetransmits is not None and not isinstance(params.maxRetransmits, int):
        raise TypeError("invalid params.maxRetransmits")

    if params.maxPacketLifeTime is not None and params.maxRetransmits is not None:
        raise TypeError("cannot provide both maxPacketLifeTime and maxRetransmits")

    if params.maxPacketLifeTime is not None or params.maxRetransmits is not None:
        params.ordered = False

    if params.label is not None and not isinstance(params.label, str):
        raise TypeError("invalid params.label")

    if params.protocol is not None and not isinstance(params.protocol, str):
        raise TypeError("invalid params.protocol")


def validateSctpCapabilities(caps: SctpCapabilities) -> None:
    if not isinstance(caps, SctpCapabilities):
        raise TypeError("caps is not an object")

    if not isinstance(caps.numStreams, NumSctpStreams):
        raise TypeError("missing caps.numStreams")

    validateNumSctpStreams(caps.numStreams)


def validateNumSctpStreams(numStreams: NumSctpStreams) -> None:
    if not isinstance(numStreams, NumSctpStreams):
        raise TypeError("numStreams is not an object")

    if not isinstance(numStreams.OS, int):
        raise TypeError("missing numStreams.OS")

    if not isinstance(numStreams.MIS, int):
        raise TypeError("missing numStreams.MIS")


def validateAndNormalizeRtpCodecCapability(codec: RtpCodecCapability) -> None:
    if not isinstance(codec, RtpCodecCapability):
        raise TypeError("codec is not an object")

    mimeTypeParts = codec.mimeType.split("/", 1)
    if len(mimeTypeParts) != 2:
        raise TypeError("invalid codec.mimeType")

    kind = mimeTypeParts[0].lower()
    if kind not in ["audio", "video"]:
        raise TypeError("invalid codec.mimeType")
    codec.kind = cast(MediaKind, kind)

    if codec.preferredPayloadType is None or not isinstance(
        codec.preferredPayloadType, int
    ):
        raise TypeError("missing codec.preferredPayloadType")

    if not isinstance(codec.clockRate, int):
        raise TypeError("missing codec.clockRate")

    if codec.kind == "audio":
        if not isinstance(codec.channels, int):
            codec.channels = 1
    else:
        codec.channels = None

    _normalizeCodecParameters(codec)
    _normalizeRtcpFeedbackList(codec.rtcpFeedback)


def validateAndNormalizeRtpCodecParameters(codec: RtpCodecParameters) -> None:
    if not isinstance(codec, RtpCodecParameters):
        raise TypeError("codec is not an object")

    mimeTypeParts = codec.mimeType.split("/", 1)
    if len(mimeTypeParts) != 2:
        raise TypeError("invalid codec.mimeType")

    if not isinstance(codec.payloadType, int):
        raise TypeError("missing codec.payloadType")

    if not isinstance(codec.clockRate, int):
        raise TypeError("missing codec.clockRate")

    kind = mimeTypeParts[0].lower()
    if kind == "audio":
        if not isinstance(codec.channels, int):
            codec.channels = 1
    else:
        codec.channels = None

    _normalizeCodecParameters(codec)
    _normalizeRtcpFeedbackList(codec.rtcpFeedback)


def validateAndNormalizeRtpHeaderExtension(ext: RtpHeaderExtension) -> None:
    if not isinstance(ext, RtpHeaderExtension):
        raise TypeError("ext is not an object")

    if ext.kind is not None and ext.kind not in ["audio", "video"]:
        raise TypeError("invalid ext.kind")

    if not isinstance(ext.uri, str):
        raise TypeError("missing ext.uri")

    if not isinstance(ext.preferredId, int):
        raise TypeError("missing ext.preferredId")

    if ext.preferredEncrypt is None:
        ext.preferredEncrypt = False
    elif not isinstance(ext.preferredEncrypt, bool):
        raise TypeError("invalid ext.preferredEncrypt")

    if ext.direction is None:
        ext.direction = "sendrecv"
    elif ext.direction not in ["sendrecv", "sendonly", "recvonly", "inactive"]:
        raise TypeError("invalid ext.direction")


def validateRtpHeaderExtensionParameters(ext: RtpHeaderExtensionParameters) -> None:
    if not isinstance(ext, RtpHeaderExtensionParameters):
        raise TypeError("ext is not an object")

    if not isinstance(ext.uri, str):
        raise TypeError("missing ext.uri")

    if not isinstance(ext.id, int):
        raise TypeError("missing ext.id")

    if ext.encrypt is None:
        ext.encrypt = False
    elif not isinstance(ext.encrypt, bool):
        raise TypeError("invalid ext.encrypt")

    if ext.parameters is None:
        ext.parameters = {}
    elif not isinstance(ext.parameters, dict):
        raise TypeError("invalid ext.parameters")

    for key, value in list(ext.parameters.items()):
        if value is None:
            ext.parameters[key] = ""
            value = ""

        if not isinstance(value, (str, int, float)):
            raise TypeError("invalid header extension parameter")


def validateAndNormalizeRtpEncodingParameters(encoding: RtpEncodingParameters) -> None:
    if not isinstance(encoding, RtpEncodingParameters):
        raise TypeError("encoding is not an object")

    if encoding.ssrc is not None and not isinstance(encoding.ssrc, int):
        raise TypeError("invalid encoding.ssrc")

    if encoding.rid is not None and not isinstance(encoding.rid, str):
        raise TypeError("invalid encoding.rid")

    if encoding.codecPayloadType is not None and not isinstance(
        encoding.codecPayloadType, int
    ):
        raise TypeError("invalid encoding.codecPayloadType")

    if encoding.rtx is not None and not isinstance(encoding.rtx.ssrc, int):
        raise TypeError("missing encoding.rtx.ssrc")

    if not isinstance(encoding.dtx, bool):
        encoding.dtx = False

    if encoding.scalabilityMode is not None and not isinstance(
        encoding.scalabilityMode, str
    ):
        raise TypeError("invalid encoding.scalabilityMode")


def validateAndNormalizeRtcpParameters(rtcp: RtcpParameters) -> None:
    if not isinstance(rtcp, RtcpParameters):
        raise TypeError("rtcp is not an object")

    if rtcp.cname is not None and not isinstance(rtcp.cname, str):
        raise TypeError("invalid rtcp.cname")

    if not isinstance(rtcp.reducedSize, bool):
        rtcp.reducedSize = True


def _getRtpCapabilities(
    direction: str, extendedRtpCapabilities: ExtendedRtpCapabilities
) -> RtpCapabilities:
    rtpCapabilities = RtpCapabilities()

    for extendedCodec in extendedRtpCapabilities.codecs:
        if direction == "recvonly":
            preferredPayloadType = extendedCodec.remotePayloadType
            rtxPayloadType = extendedCodec.remoteRtxPayloadType
            apt = extendedCodec.remotePayloadType
        else:
            preferredPayloadType = extendedCodec.localPayloadType
            rtxPayloadType = extendedCodec.localRtxPayloadType
            apt = extendedCodec.localPayloadType

        rtpCapabilities.codecs.append(
            RtpCodecCapability(
                kind=extendedCodec.kind,
                mimeType=extendedCodec.mimeType,
                preferredPayloadType=preferredPayloadType,
                clockRate=extendedCodec.clockRate,
                channels=extendedCodec.channels,
                parameters=extendedCodec.localParameters,
                rtcpFeedback=extendedCodec.rtcpFeedback,
            )
        )

        if rtxPayloadType is not None:
            rtpCapabilities.codecs.append(
                RtpCodecCapability(
                    kind=extendedCodec.kind,
                    mimeType=f"{extendedCodec.kind}/rtx",
                    preferredPayloadType=rtxPayloadType,
                    clockRate=extendedCodec.clockRate,
                    parameters={"apt": apt},
                    rtcpFeedback=[],
                )
            )

    for extendedExtension in extendedRtpCapabilities.headerExtensions:
        if extendedExtension.direction not in ["sendrecv", direction]:
            continue

        preferredId = (
            extendedExtension.recvId
            if direction == "recvonly"
            else extendedExtension.sendId
        )

        rtpCapabilities.headerExtensions.append(
            RtpHeaderExtension(
                kind=extendedExtension.kind,
                uri=extendedExtension.uri,
                preferredId=preferredId,
                preferredEncrypt=extendedExtension.encrypt,
                direction=extendedExtension.direction,
            )
        )

    return rtpCapabilities


def _normalizeCodecParameters(codec: RtpCodec) -> None:
    if codec.parameters is None:
        codec.parameters = {}
    elif not isinstance(codec.parameters, dict):
        raise TypeError("invalid codec.parameters")

    for key, value in list(codec.parameters.items()):
        if value is None:
            codec.parameters[key] = ""
            value = ""

        if not isinstance(value, (str, int, float)):
            raise TypeError(f"invalid codec parameter [key:{key}, value:{value}]")

        if key == "apt" and not isinstance(value, int):
            raise TypeError("invalid codec apt parameter")


def _normalizeRtcpFeedbackList(feedbacks: List[RtcpFeedback]) -> None:
    if not isinstance(feedbacks, list):
        raise TypeError("codec.rtcpFeedback is not an array")

    for feedback in feedbacks:
        if not isinstance(feedback.type, str):
            raise TypeError("missing fb.type")

        if feedback.parameter is None or not isinstance(feedback.parameter, str):
            feedback.parameter = ""



import logging
from typing import List, Optional, Any
from pyee.asyncio import AsyncIOEventEmitter
from pydantic import BaseModel, Field, ConfigDict
from aiortc import RTCRtpSender, MediaStreamTrack
from .emitter import EnhancedEventEmitter
from .errors import InvalidStateError, UnsupportedError
from .rtp_parameters import RtpParameters, RtpCodecCapability, RtpEncodingParameters


logger = logging.getLogger(__name__)


# https://mediasoup.org/documentation/v3/mediasoup-client/api/#ProducerCodecOptions
class ProducerCodecOptions(BaseModel):
    opusStereo: Optional[bool] = None
    opusFec: Optional[bool] = None
    opusDtx: Optional[bool] = None
    opusMaxPlaybackRate: Optional[int] = None
    opusMaxAverageBitrate: Optional[int] = None
    opusPtime: Optional[int] = None
    videoGoogleStartBitrate: Optional[int] = None
    videoGoogleMaxBitrate: Optional[int] = None
    videoGoogleMinBitrate: Optional[int] = None


class ProducerOptions(BaseModel):
    track: Optional[MediaStreamTrack] = None
    encodings: Optional[List[RtpEncodingParameters]] = Field(default_factory=list)
    codecOptions: Optional[ProducerCodecOptions] = None
    codec: Optional[RtpCodecCapability] = None
    streamId: Optional[str] = None
    headerExtensionOptions: Optional[dict] = None
    stopTracks: bool = True
    disableTrackOnPause: bool = True
    zeroRtpOnPause: bool = False
    appData: Optional[Any] = Field(default_factory=dict)

    model_config = ConfigDict(arbitrary_types_allowed=True)


class Producer(EnhancedEventEmitter):
    def __init__(
        self,
        id: str,
        localId: str,
        track: MediaStreamTrack,
        rtpParameters: RtpParameters,
        stopTracks: bool,
        disableTrackOnPause: bool,
        zeroRtpOnPause: bool,
        rtpSender: Optional[RTCRtpSender] = None,
        appData: Optional[dict] = None,
        loop=None,
    ):
        super(Producer, self).__init__(loop=loop)

        # Closed flag.
        self._closed: bool = False
        # Observer instance.
        self._observer: AsyncIOEventEmitter = AsyncIOEventEmitter()

        self._id = id
        self._localId = localId
        self._rtpSender = rtpSender
        self._track = track
        self._kind = track.kind
        self._rtpParameters = rtpParameters
        self._paused = bool(disableTrackOnPause and not getattr(track, "enabled", True))
        self._maxSpatialLayer: Optional[int] = None
        self._stopTracks = stopTracks
        self._disableTrackOnPause = disableTrackOnPause
        self._zeroRtpOnPause = zeroRtpOnPause
        self._appData = appData if appData is not None else {}

        self._handleTrack()

    # Producer id.
    @property
    def id(self) -> str:
        return self._id

    # Local id.
    @property
    def localId(self) -> str:
        return self._localId

    # Whether the Producer is closed.
    @property
    def closed(self) -> bool:
        return self._closed

    # Media kind.
    @property
    def kind(self) -> str:
        return self._kind

    # Associated RTCRtpSender.
    @property
    def rtpSender(self) -> Optional[RTCRtpSender]:
        return self._rtpSender

    # The associated track.
    @property
    def track(self) -> Optional[MediaStreamTrack]:
        return self._track

    # RTP parameters.
    @property
    def rtpParameters(self) -> RtpParameters:
        return self._rtpParameters

    # Whether the Producer is paused.
    @property
    def paused(self) -> bool:
        return self._paused

    # Max spatial layer.
    @property
    def maxSpatialLayer(self) -> Optional[int]:
        return self._maxSpatialLayer

    # App custom data.
    @property
    def appData(self) -> Any:
        return self._appData

    # Invalid setter.
    @appData.setter
    def appData(self, value):
        self._appData = value

    # Observer.
    #
    # @emits close
    # @emits pause
    # @emits resume
    # @emits trackended
    @property
    def observer(self) -> AsyncIOEventEmitter:
        return self._observer

    async def close(self):
        if self._closed:
            return

        logger.debug("Producer close()")

        self._closed = True

        self._destroyTrack()

        await self.emit_for_results("@close")

        # Emit observer event.
        self._observer.emit("close")

    # Transport was closed.
    def transportClosed(self):
        if self._closed:
            return

        logger.debug("Producer transportClosed()")

        self._closed = True

        self._destroyTrack()

        self.emit("transportclose")

        self._observer.emit("close")

    # Get associated RTCRtpSender stats.
    async def getStats(self):
        if self._closed:
            raise InvalidStateError("closed")

        return await self.emit_for_results("@getstats")

    # Pauses sending media.
    def pause(self):
        logger.debug("Producer pause()")

        if self._closed:
            logger.debug("Producer pause() | Producer closed")
            return

        if self._paused:
            logger.debug("Producer pause() | Producer already paused")
            return

        self._paused = True

        self._syncTrackEnabled()

        if self._zeroRtpOnPause:
            self.emit("@pause")

        self._observer.emit("pause")

    # Resumes sending media.
    def resume(self):
        logger.debug("Producer resume()")

        if self._closed:
            logger.debug("Producer resume() | Producer closed")
            return

        if not self._paused:
            logger.debug("Producer resume() | Producer already resumed")
            return

        self._paused = False

        self._syncTrackEnabled()

        if self._zeroRtpOnPause:
            self.emit("@resume")

        self._observer.emit("resume")

    # Replaces the current track with a new one or null.
    async def replaceTrack(self, track: Optional[MediaStreamTrack] = None):
        logger.debug(f"replaceTrack() [track: {track}]")

        if self._closed:
            # This must be done here. Otherwise there is no chance to stop the given
            # track.
            if track and self._stopTracks:
                track.stop()

            raise InvalidStateError("closed")

        elif track and track.readyState == "ended":
            raise InvalidStateError("track ended")

        # Do nothing if this is the same track as the current handled one.
        if track == self._track:
            logger.debug("Producer replaceTrack() | same track, ignored")
            return

        if not self._zeroRtpOnPause or not self._paused:
            await self.emit_for_results("@replacetrack", track)

        # Destroy the previous track.
        self._destroyTrack()

        self._track = track

        # If this Producer was paused/resumed and the state of the new
        # track does not match, fix it.
        self._syncTrackEnabled()

        self._handleTrack()

    # Sets the video max spatial layer to be sent.
    async def setMaxSpatialLayer(self, spatialLayer: int):
        if self._closed:
            raise InvalidStateError("closed")

        elif self.kind != "video":
            raise UnsupportedError("not a video Producer")

        elif not isinstance(spatialLayer, int):
            raise TypeError("invalid spatialLayer")

        if spatialLayer == self._maxSpatialLayer:
            return

        await self.emit_for_results("@setmaxspatiallayer", spatialLayer)

        self._maxSpatialLayer = spatialLayer

    # Sets the DSCP value.
    # TODO: RTCRtpEncodingParameters
    async def setRtpEncodingParameters(self, params):
        if self._closed:
            raise InvalidStateError("closed")

        await self.emit_for_results("@setrtpencodingparameters", params)

    def _onTrackEnded(self):
        logger.debug('Producer track "ended" event')
        self.emit("trackended")
        self._observer.emit("trackended")

    def _handleTrack(self):
        if not self._track:
            return

        self._track.on("ended", self._onTrackEnded)

    def _syncTrackEnabled(self):
        if not self._track or not self._disableTrackOnPause:
            return

        if hasattr(self._track, "enabled"):
            setattr(self._track, "enabled", not self._paused)

    def _destroyTrack(self):
        if not self._track:
            return

        try:
            self._track.remove_listener("ended", self._onTrackEnded)
        except Exception:
            pass

        if self._track.readyState != "ended" and self._stopTracks:
            self._track.stop()



from typing import Optional, Literal, List, Any, Dict

import logging
from copy import deepcopy
from pyee.asyncio import AsyncIOEventEmitter
from aiortc import RTCIceServer, MediaStreamTrack
from .ortc import (
    canReceive,
    generateProbatorRtpParameters,
    ExtendedRtpCapabilities,
    validateAndNormalizeSctpStreamParameters,
)
from .errors import InvalidStateError, UnsupportedError
from .emitter import EnhancedEventEmitter
from .handlers.handler_interface import HandlerInterface
from .models.handler_interface import (
    HandlerSendResult,
    HandlerReceiveResult,
    SctpStreamParameters,
    HandlerSendDataChannelResult,
    HandlerReceiveDataChannelResult,
)
from .models.transport import (
    ConnectionState,
    IceParameters,
    InternalTransportOptions,
    DtlsParameters,
)
from .consumer import Consumer, ConsumerOptions
from .producer import Producer, ProducerOptions
from .data_consumer import DataConsumer, DataConsumerOptions
from .data_producer import DataProducer, DataProducerOptions
from .producer import ProducerCodecOptions
from .rtp_parameters import (
    RtpParameters,
    RtpCodecCapability,
    RtpEncodingParameters,
    MediaKind,
)


logger = logging.getLogger(__name__)


class Transport(EnhancedEventEmitter):
    def __init__(self, options: InternalTransportOptions, loop=None):
        super(Transport, self).__init__(loop=loop)

        logger.debug(f"constructor() [id:{options.id}, direction:{options.direction}]")

        # Closed flag.
        self._closed: bool = False
        # Whether we can produce audio/video based on computed extended RTP
        # capabilities.
        self._canProduceByKind: Dict[str, bool]
        # App custom data.
        self._appData: Optional[dict]
        # Transport connection state.
        self._connectionState: ConnectionState = "new"
        # Producers indexed by id
        self._producers: Dict[str, Producer] = {}
        # Consumers indexed by id.
        self._consumers: Dict[str, Consumer] = {}
        # DataProducers indexed by id
        self._dataProducers: Dict[str, DataProducer] = {}
        # DataConsumers indexed by id.
        self._dataConsumers: Dict[str, DataConsumer] = {}
        # Whether the Consumer for RTP probation has been created.
        self._probatorConsumerCreated: bool = False
        # Observer instance.
        self._observer: AsyncIOEventEmitter = AsyncIOEventEmitter()

        # Id.
        self._id: str = options.id
        # Direction.
        self._direction: Literal["send", "recv"] = options.direction
        # Extended RTP capabilities.
        extendedRtpCapabilities = options.extendedRtpCapabilities
        if extendedRtpCapabilities is None:
            raise TypeError("missing extendedRtpCapabilities")
        self._extendedRtpCapabilities: ExtendedRtpCapabilities = extendedRtpCapabilities
        self._canProduceByKind: Dict[str, bool] = options.canProduceByKind
        self._maxSctpMessageSize = (
            options.sctpParameters.maxMessageSize if options.sctpParameters else None
        )

        if options.additionalSettings:
            additionalSettings = deepcopy(options.additionalSettings)
            additionalSettings.pop("iceServers", None)
            additionalSettings.pop("iceTransportPolicy", None)
            additionalSettings.pop("bundlePolicy", None)
            additionalSettings.pop("rtcpMuxPolicy", None)
            additionalSettings.pop("sdpSemantics", None)
        else:
            additionalSettings = None

        self._handler: HandlerInterface = options.handlerFactory()

        self._handler.run(
            direction=options.direction,
            iceParameters=options.iceParameters,
            iceCandidates=options.iceCandidates,
            dtlsParameters=options.dtlsParameters,
            sctpParameters=options.sctpParameters,
            iceServers=options.iceServers,
            iceTransportPolicy=options.iceTransportPolicy,
            additionalSettings=additionalSettings,
            proprietaryConstraints=options.proprietaryConstraints,
            extendedRtpCapabilities=extendedRtpCapabilities,
        )

        self._appData = options.appData

        self._handleHandler()

    # Producer id.
    @property
    def id(self) -> str:
        return self._id

    # Whether the Producer is closed.
    @property
    def closed(self) -> bool:
        return self._closed

    # Transport direction.
    @property
    def direction(self) -> Literal["send", "recv"]:
        return self._direction

    # RTC handler instance.
    @property
    def handler(self) -> HandlerInterface:
        return self._handler

    # Connection state.
    @property
    def connectionState(self) -> ConnectionState:
        return self._connectionState

    # App custom data.
    @property
    def appData(self) -> Any:
        return self._appData

    # App custom data setter.
    @appData.setter
    def appData(self, value):
        self._appData = value

    # Observer.
    #
    # @emits close
    # @emits newproducer - (producer: Producer)
    # @emits newconsumer - (producer: Producer)
    # @emits newdataproducer - (dataProducer: DataProducer)
    # @emits newdataconsumer - (dataProducer: DataProducer)
    @property
    def observer(self) -> AsyncIOEventEmitter:
        return self._observer

    # Close the Transport.
    async def close(self):
        if self._closed:
            return

        logger.debug("Transport close()")

        self._closed = True
        self._connectionState = "closed"

        # Close the handler.
        await self._handler.close()

        # Close all Process.
        for process_dict in [
            self._producers,
            self._consumers,
            self._dataProducers,
            self._dataConsumers,
        ]:
            for process in process_dict.values():
                process.transportClosed()
            process_dict.clear()

        self._observer.emit("close")

    # Get associated Transport (RTCPeerConnection) stats.
    #
    # @returns {RTCStatsReport}
    async def getStats(self):
        if self._closed:
            raise InvalidStateError("closed")

        return await self._handler.getTransportStats()

    # Restart ICE connection.
    async def restartIce(self, iceParameters: IceParameters):
        if self._closed:
            raise InvalidStateError("closed")

        return await self._handler.restartIce(iceParameters)

    # Update ICE servers.
    async def updateIceServers(self, iceServers: List[RTCIceServer]):
        if self._closed:
            raise InvalidStateError("closed")

        return await self._handler.updateIceServers(iceServers)

    # Create a Producer.
    async def produce(
        self,
        track: Optional[MediaStreamTrack] = None,
        encodings: Optional[List[RtpEncodingParameters]] = None,
        codecOptions: Optional[ProducerCodecOptions] = None,
        codec: Optional[RtpCodecCapability] = None,
        streamId: Optional[str] = None,
        headerExtensionOptions: Optional[dict] = None,
        stopTracks: bool = True,
        disableTrackOnPause: bool = True,
        zeroRtpOnPause: bool = False,
        appData: Optional[Any] = None,
    ) -> Producer:
        options: ProducerOptions = ProducerOptions(
            track=track,
            encodings=encodings,
            codecOptions=codecOptions,
            codec=codec,
            streamId=streamId,
            headerExtensionOptions=headerExtensionOptions,
            stopTracks=stopTracks,
            disableTrackOnPause=disableTrackOnPause,
            zeroRtpOnPause=zeroRtpOnPause,
            appData=appData if appData is not None else {},
        )
        logger.debug(f"Transport produce() [track:{options.track}]")
        if self._closed:
            raise InvalidStateError("closed")
        elif not options.track:
            raise TypeError("missing track")
        elif self._direction != "send":
            raise UnsupportedError("not a sending Transport")
        elif not self._canProduceByKind.get(options.track.kind):
            raise UnsupportedError(f"cannot produce {options.track.kind}")
        elif options.track.readyState == "ended":
            raise InvalidStateError("track ended")
        elif len(self.listeners("connect")) == 0 and self._connectionState == "new":
            raise TypeError('no "connect" listener set into this transport')
        elif len(self.listeners("produce")) == 0:
            raise TypeError('no "produce" listener set into this transport')

        # NOTE: Mediasoup client enqueue command here.
        handlerSendResult: HandlerSendResult = await self._handler.send(
            track=options.track,
            encodings=options.encodings,
            codecOptions=options.codecOptions,
            codec=options.codec,
            streamId=options.streamId,
            headerExtensionOptions=options.headerExtensionOptions,
        )

        ids = await self.emit_for_results(
            "produce",
            options.track.kind,
            handlerSendResult.rtpParameters,
            options.appData,
        )

        producer = Producer(
            id=ids[0],
            localId=handlerSendResult.localId,
            rtpSender=handlerSendResult.rtpSender,
            track=options.track,
            rtpParameters=handlerSendResult.rtpParameters,
            stopTracks=options.stopTracks,
            disableTrackOnPause=options.disableTrackOnPause,
            zeroRtpOnPause=options.zeroRtpOnPause,
            appData=options.appData,
        )

        self._producers[producer.id] = producer
        self._handleProducer(producer)

        # Emit observer event.
        self._observer.emit("newproducer", producer)

        return producer

        # TODO: stop the given track if the command above failed due to closed Transport.

    async def consume(
        self,
        id: str,
        producerId: str,
        kind: MediaKind,
        rtpParameters: RtpParameters,
        streamId: Optional[str] = None,
        appData: Optional[dict] = None,
    ) -> Consumer:
        options: ConsumerOptions = ConsumerOptions(
            id=id,
            producerId=producerId,
            kind=kind,
            rtpParameters=rtpParameters,
            streamId=streamId,
            appData=appData if appData is not None else {},
        )
        logger.debug("Transport consume()")
        rtpParameters = options.rtpParameters.model_copy(deep=True)
        if self._closed:
            raise InvalidStateError("closed")
        elif self._direction != "recv":
            raise UnsupportedError("not a receiving Transport")
        elif len(self.listeners("connect")) == 0 and self._connectionState == "new":
            raise TypeError('no "connect" listener set into this transport')

        # NOTE: Mediasoup client enqueue command here.
        if not canReceive(
            rtpParameters=rtpParameters,
            extendedRtpCapabilities=self._extendedRtpCapabilities,
        ):
            raise UnsupportedError("cannot consume this Producer")

        handlerReceiveResult: HandlerReceiveResult = await self._handler.receive(
            trackId=options.id,
            kind=options.kind,
            rtpParameters=rtpParameters,
            streamId=options.streamId,
        )

        consumer: Consumer = Consumer(
            id=options.id,
            localId=handlerReceiveResult.localId,
            producerId=options.producerId,
            track=handlerReceiveResult.track,
            rtpParameters=rtpParameters,
            appData=options.appData,
        )

        self._consumers[consumer.id] = consumer
        self._handleConsumer(consumer)

        # If this is the first video Consumer and the Consumer for RTP probation
        # has not yet been created, create it now.
        if not self._probatorConsumerCreated and options.kind == "video":
            probatorRtpParameters = generateProbatorRtpParameters(
                consumer.rtpParameters
            )
            await self._handler.receive(
                trackId="probator", kind="video", rtpParameters=probatorRtpParameters
            )

            logger.debug("Transport consume() | Consumer for RTP probation created")

            self._probatorConsumerCreated = True

        self._observer.emit("newconsumer", consumer)

        return consumer

    # Create a DataProducer
    async def produceData(
        self,
        ordered: Optional[bool] = True,
        maxPacketLifeTime: Optional[int] = None,
        maxRetransmits: Optional[int] = None,
        label: Optional[str] = None,
        protocol: Optional[str] = None,
        appData: Optional[dict] = None,
    ) -> DataProducer:
        options: DataProducerOptions = DataProducerOptions(
            ordered=ordered,
            maxPacketLifeTime=maxPacketLifeTime,
            maxRetransmits=maxRetransmits,
            label=label,
            protocol=protocol,
            appData=appData if appData is not None else {},
        )
        logger.debug("Transport produceData()")
        if self._closed:
            raise InvalidStateError("closed")
        elif self._direction != "send":
            raise UnsupportedError("not a sending Transport")

        elif not self._maxSctpMessageSize:
            raise UnsupportedError("SCTP not enabled by remote Transport")

        elif len(self.listeners("connect")) == 0 and self._connectionState == "new":
            raise TypeError('no "connect" listener set into this transport')

        elif len(self.listeners("producedata")) == 0:
            raise TypeError('no "producedata" listener set into this transport')

        if options.maxPacketLifeTime or options.maxRetransmits:
            options.ordered = False

        # NOTE: Mediasoup client enqueue command here.
        handlerSendDataChannelResult: HandlerSendDataChannelResult = (
            await self._handler.sendDataChannel(
                ordered=options.ordered,
                maxPacketLifeTime=options.maxPacketLifeTime,
                maxRetransmits=options.maxRetransmits,
                label=options.label,
                protocol=options.protocol,
            )
        )

        ids = await self.emit_for_results(
            "producedata",
            sctpStreamParameters=handlerSendDataChannelResult.sctpStreamParameters,
            label=options.label,
            protocol=options.protocol,
            appData=options.appData,
        )

        dataProducer = DataProducer(
            id=ids[0],
            dataChannel=handlerSendDataChannelResult.dataChannel,
            sctpStreamParameters=handlerSendDataChannelResult.sctpStreamParameters,
            appData=options.appData,
        )

        self._dataProducers[dataProducer.id] = dataProducer
        self._handleDataProducer(dataProducer)

        # Emit observer event.
        self._observer.emit("newdataproducer", dataProducer)

        return dataProducer

    # Create a DataConsumer
    async def consumeData(
        self,
        id: str,
        dataProducerId: str,
        sctpStreamParameters: SctpStreamParameters,
        label: Optional[str] = None,
        protocol: Optional[str] = None,
        appData: Optional[dict] = None,
    ) -> DataConsumer:
        options: DataConsumerOptions = DataConsumerOptions(
            id=id,
            dataProducerId=dataProducerId,
            sctpStreamParameters=sctpStreamParameters,
            label=label,
            protocol=protocol,
            appData=appData if appData is not None else {},
        )
        logger.debug("Transport consumeData()")
        validateAndNormalizeSctpStreamParameters(options.sctpStreamParameters)
        if self._closed:
            raise InvalidStateError("closed")
        elif self._direction != "recv":
            raise UnsupportedError("not a receiving Transport")
        elif not self._maxSctpMessageSize:
            raise UnsupportedError("SCTP not enabled by remote Transport")
        elif len(self.listeners("connect")) == 0 and self._connectionState == "new":
            raise TypeError('no "connect" listener set into this transport')

        # NOTE: Mediasoup client enqueue command here.
        handlerReceiveDataChannelResult: HandlerReceiveDataChannelResult = (
            await self._handler.receiveDataChannel(
                sctpStreamParameters=options.sctpStreamParameters,
                label=options.label,
                protocol=options.protocol,
            )
        )

        dataConsumer: DataConsumer = DataConsumer(
            id=options.id,
            dataProducerId=options.dataProducerId,
            dataChannel=handlerReceiveDataChannelResult.dataChannel,
            sctpStreamParameters=options.sctpStreamParameters,
            appData=options.appData,
        )

        self._dataConsumers[dataConsumer.id] = dataConsumer
        self._handleDataConsumer(dataConsumer)

        # Emit observer event.
        self._observer.emit("newdataconsumer", dataConsumer)

        return dataConsumer

    def _handleHandler(self):
        handler = self._handler

        @handler.on("@connect")
        async def on_connect(dtlsParameters: DtlsParameters):
            if self._closed:
                raise InvalidStateError("closed")
            else:
                self.emit("connect", dtlsParameters)

        @handler.on("@connectionstatechange")
        def on_connectionstatechange(connectionState: ConnectionState):
            self._connectionState = connectionState
            if not self._closed:
                self.emit("connectionstatechange", connectionState)

    def _handleProducer(self, producer: Producer):
        @producer.on("@close")
        async def on_close():
            del self._producers[producer.id]
            if self._closed:
                return
            await self._handler.stopSending(producer.localId)

        @producer.on("@pause")
        async def on_pause():
            await self._handler.pauseSending(producer.localId)

        @producer.on("@resume")
        async def on_resume():
            await self._handler.resumeSending(producer.localId)

        @producer.on("@replacetrack")
        async def on_replacetrack(track):
            await self._handler.replaceTrack(producer.localId, track)

        @producer.on("@setmaxspatiallayer")
        async def on_setmaxspatiallayer(spatialLayer):
            await self._handler.setMaxSpatialLayer(producer.localId, spatialLayer)

        @producer.on("@setrtpencodingparameters")
        async def on_setrtpencodingparameters(params):
            await self._handler.setRtpEncodingParameters(producer.localId, params)

        @producer.on("@getstats")
        async def on_getstats():
            if self._closed:
                return InvalidStateError("closed")
            return await self._handler.getSenderStats(producer.localId)

    def _handleConsumer(self, consumer: Consumer):
        @consumer.on("@close")
        async def on_close():
            del self._consumers[consumer.id]
            if self._closed:
                return
            await self._handler.stopReceiving(consumer.localId)

        @consumer.on("@pause")
        async def on_pause():
            await self._handler.pauseReceiving(consumer.localId)

        @consumer.on("@resume")
        async def on_resume():
            await self._handler.resumeReceiving(consumer.localId)

        @consumer.on("@getstats")
        async def on_getstats():
            if self._closed:
                return InvalidStateError("closed")
            return await self._handler.getReceiverStats(consumer.localId)

    def _handleDataProducer(self, dataProducer: DataProducer):
        @dataProducer.on("@close")
        async def on_close():
            del self._dataProducers[dataProducer.id]

    def _handleDataConsumer(self, dataConsumer: DataConsumer):
        @dataConsumer.on("@close")
        async def on_close():
            del self._dataConsumers[dataConsumer.id]



from typing import Literal, List, Optional, Any

from pydantic import BaseModel, Field, ConfigDict
from aiortc import (
    RTCIceServer,
    RTCRtpSender,
    RTCRtpReceiver,
    RTCDataChannel,
    MediaStreamTrack,
)
from .transport import IceCandidate, IceParameters, DtlsParameters
from ..ortc import ExtendedRtpCapabilities
from ..sctp_parameters import SctpParameters, SctpStreamParameters
from ..producer import ProducerCodecOptions
from ..rtp_parameters import (
    RtpCodecCapability,
    RtpParameters,
    MediaKind,
    RtpEncodingParameters,
)


class HandlerRunOptions(BaseModel):
    direction: Literal["send", "recv"]
    iceParameters: IceParameters
    iceCandidates: List[IceCandidate]
    dtlsParameters: DtlsParameters
    sctpParameters: Optional[SctpParameters] = None
    iceServers: Optional[List[RTCIceServer]]
    iceTransportPolicy: Optional[Literal["all", "relay"]]
    additionalSettings: Optional[Any] = None
    proprietaryConstraints: Optional[Any] = None
    extendedRtpCapabilities: ExtendedRtpCapabilities

    model_config = ConfigDict(arbitrary_types_allowed=True)


class HandlerSendOptions(BaseModel):
    track: MediaStreamTrack
    encodings: List[RtpEncodingParameters] = Field(default_factory=list)
    codecOptions: Optional[ProducerCodecOptions] = None
    codec: Optional[RtpCodecCapability] = None
    streamId: Optional[str] = None
    headerExtensionOptions: Optional[dict] = None

    model_config = ConfigDict(arbitrary_types_allowed=True)


class HandlerSendResult(BaseModel):
    localId: str
    rtpParameters: RtpParameters
    rtpSender: Optional[RTCRtpSender] = None

    model_config = ConfigDict(arbitrary_types_allowed=True)


class HandlerReceiveOptions(BaseModel):
    trackId: str
    kind: MediaKind
    rtpParameters: RtpParameters
    streamId: Optional[str] = None


class HandlerReceiveResult(BaseModel):
    localId: str
    track: MediaStreamTrack
    rtpReceiver: Optional[RTCRtpReceiver] = None

    model_config = ConfigDict(arbitrary_types_allowed=True)


class HandlerSendDataChannelResult(BaseModel):
    dataChannel: RTCDataChannel
    sctpStreamParameters: SctpStreamParameters

    model_config = ConfigDict(arbitrary_types_allowed=True)


class HandlerReceiveDataChannelOptions(BaseModel):
    sctpStreamParameters: SctpStreamParameters
    label: Optional[str] = None
    protocol: Optional[str] = None


class HandlerReceiveDataChannelResult(BaseModel):
    dataChannel: RTCDataChannel

    model_config = ConfigDict(arbitrary_types_allowed=True)
